package dev.fitko.fitconnect.client;

import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.SubmissionRequestException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.client.subscriber.SubmissionReceiver;
import dev.fitko.fitconnect.client.subscriber.model.DecryptedAttachmentPayload;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Set;
import java.util.UUID;

/**
 * A subscriber-side client for retrieving submissions.
 */
public class SubscriberClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubscriberClient.class);

    private static final int DEFAULT_SUBMISSION_LIMIT = 500;

    private final Subscriber subscriber;
    private final SubmissionReceiver submissionReceiver;

    public SubscriberClient(final Subscriber subscriber, final SubmissionReceiver submissionReceiver) {
        this.subscriber = subscriber;
        this.submissionReceiver = submissionReceiver;
    }

    /**
     * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}s destination.
     *
     * @param destinationId filter criterion for the submissions
     * @return set of the first 0..500 available submissions
     */
    public Set<SubmissionForPickup> getAvailableSubmissionsForDestination(final UUID destinationId) {
        return getAvailableSubmissionsForDestination(destinationId, 0, DEFAULT_SUBMISSION_LIMIT);
    }

    /**
     * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}s destination.
     *
     * @param destinationId filter criterion for the submissions
     * @param offset        position in the dataset
     * @param limit         number of submissions in result (max. is 500)
     * @return set of available submissions in the given range
     */
    public Set<SubmissionForPickup> getAvailableSubmissionsForDestination(final UUID destinationId, final int offset, final int limit) {
        final Set<SubmissionForPickup> submissions = subscriber.pollAvailableSubmissionsForDestination(destinationId, offset, limit);
        LOGGER.info("Received {} submission(s) for destination {}", submissions.size(), destinationId);
        return submissions;
    }

    /**
     * Retrieve the entire event log for a submissions caseId of a specific destination.
     *
     * @param caseId        unique identifier of the case the log should be retrieved for
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     * @return List of {@link EventLogEntry}s for the given case
     */
    public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) {
        return subscriber.getEventLog(caseId, destinationId);
    }

    /**
     * Loads a single {@link SubmissionForPickup} by id.
     *
     * @param submission {@link SentSubmission} that is requested
     * @return {@link ReceivedSubmission} to get the submission {@link Metadata}, {@link Data} and {@link ApiAttachment}s as well as
     * accept or reject the loaded submission
     */
    public ReceivedSubmission requestSubmission(final SubmissionForPickup submission) {
        return requestSubmission(submission.getSubmissionId());
    }

    /**
     * Loads a single {@link SubmissionForPickup} by id. Auto-rejects invalid submissions where e.g. a validation failed.
     *
     * @param submissionId unique identifier of the requested submission
     * @return {@link ReceivedSubmission} to get the submission {@link Metadata}, {@link Data} and {@link Attachment}s as well as accept or reject the loaded submission
     * @throws SubmissionRequestException if a technical error occurred or validation failed
     */
    public ReceivedSubmission requestSubmission(final UUID submissionId) throws SubmissionRequestException {
        return submissionReceiver.receiveSubmission(submissionId);
    }

    /**
     * Send a reject-submission {@link Event} to confirm that the handed in submission was rejected.
     *
     * @param submissionForPickup the submission that should be rejected
     * @param rejectionProblems   list of {@link Problem}s that give more detailed information on why the submission was rejected
     * @see <a href="https://docs.fitko.de/fit-connect/docs/getting-started/event-log/events">FIT-Connect Events</a>
     */
    public void rejectSubmission(final SubmissionForPickup submissionForPickup, final List<Problem> rejectionProblems) {
        subscriber.rejectSubmission(EventPayload.forRejectEvent(submissionForPickup, rejectionProblems));
    }

    /**
     * Checks if a received callback can be trusted by validating the provided request data
     *
     * @param hmac           authentication code provided by the callback
     * @param timestamp      timestamp provided by the callback
     * @param httpBody       HTTP body provided by the callback
     * @param callbackSecret secret owned by the client, which is used to calculate the hmac
     * @return {@code true} if hmac and timestamp provided by the callback meet the required conditions
     */
    public ValidationResult validateCallback(final String hmac, final Long timestamp, final String httpBody, final String callbackSecret) {
        return subscriber.validateCallback(hmac, timestamp, httpBody, callbackSecret);
    }
}
